/**CHeader*********************************************************************

  FileName    [bddfsm.h]

  PackageName [fsm]

  Synopsis    [The new preliminary interface to the fsm package]

  Description [The new preliminary interface to the fsm package.]

  SeeAlso     [optional]

  Author      [Marco Pistore]

  Copyright   [This file is part of the ``simulator'' package of NuSMV
  version 2.  Copyright (C) 2003 ITC-irst and University of Trento.

  NuSMV version 2 is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2 of
  the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it> or
  email to <nusmv-users@irst.itc.it>.  Please report bugs to
  <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to
  <nusmv@irst.itc.it>.]

******************************************************************************/
#ifndef _BDDFSM_H_
#define _BDDFSM_H_

#include "util.h"
#include "utils.h"
/* begin to port to 2.0.3. */
#include "node.h"
/* end to port to 2.0.3. */
#include "dd.h"
#include "set.h"
/* begin to port to 2.0.3. */
#include "rbc.h"
/* end to port to 2.0.3. */
#include "compile.h"
#include "img.h"
/* begin to port to 2.0.3. */
#include "opt.h"
#include "prop.h"
/* end to port to 2.0.3. */
#include "mc.h"

/* #define NuSMV_2_0_3 */

/* :TODO: remove when merging in NuSMV */
#define BddFsm Fsm_BddPtr
/* begin to port to 2.0.3. */

#ifndef NuSMV_2_0_3 
 #define CPTrans CPTrans_Ptr
#else
extern FILE * nusmv_stderr;
extern FILE * nusmv_stdout;

 #define CPTrans CP_Ptr
 #define CPList node_ptr
 #define CPTrans_Ptr CPTrans
 #define CPCluster CPCluster_Info_t
 #define CPTransGetForward CP_GetForwardTrans
 #define CPListGetItem(_a_) (CPCluster *)car(_a_)
 #define CPClusterGetTi(_a_) (_a_)->Ti
 #define CPClusterGetEi(_a_) (_a_)->Ei
 #define CPListNext(_a_) cdr(_a_)
 #define CPListIsNotEmpty(_a_) (_a_) != Nil
 #define CPListBuildSchedule compileCPBuildSchedule
#endif
/* end to port to 2.0.3. */


typedef node_ptr bdd_ptr_set;

/********************************************************************/

/* Allocation and deallocation */
BddFsm BddFsm_Alloc();
BddFsm BddFsm_Dup(BddFsm fsm);
void BddFsm_Free(BddFsm fsm);

/********************************************************************/

/* States (INVAR) */
bdd_ptr BddFsm_GetStates(BddFsm fsm);
void BddFsm_SetStates(BddFsm fsm, bdd_ptr states);

/* Fair States */
bdd_ptr BddFsm_GetFairStates(BddFsm fsm);
void BddFsm_SetFairStates(BddFsm fsm, bdd_ptr states);

/* Reachable states */
void BddFsm_ComputeReachableStates(BddFsm fsm);
bdd_ptr BddFsm_GetReachableStates(BddFsm fsm);

/* Initial states */
bdd_ptr BddFsm_GetInit(BddFsm fsm);
void BddFsm_SetInit(BddFsm fsm, bdd_ptr states);

/* Inputs */
bdd_ptr BddFsm_GetInputs(BddFsm fsm);
void BddFsm_SetInputs(BddFsm fsm, bdd_ptr inputs);

/* Transition relation */
CPTrans BddFsm_GetCPTrans(BddFsm fsm);
void BddFsm_SetCPTrans(BddFsm fsm, CPTrans cp);
bdd_ptr BddFsm_GetTrans(BddFsm fsm);

/* Fairness constraints */
bdd_ptr_set BddFsm_GetFairness(BddFsm fsm);
void BddFsm_SetFairness(BddFsm fsm, bdd_ptr_set fc);

/* Fair states */
void BddFsm_ComputeFairStates(BddFsm fsm);
bdd_ptr BddFsm_GetFairStates(BddFsm fsm);

/********************************************************************/

/* Check totality of transition relation */
/* :TODO: interface should be changed (some analysis is needed) */
int BddFsm_CheckTotal(BddFsm fsm);

/********************************************************************/

/* Shift operations */
bdd_ptr BddFsm_ShiftFwd(BddFsm fsm, bdd_ptr states);
bdd_ptr BddFsm_ShiftBwd(BddFsm fsm, bdd_ptr next_states);

/* State images (for Model Checking) */
bdd_ptr BddFsm_ImageFwd(BddFsm fsm, bdd_ptr states);
bdd_ptr BddFsm_ImageBwd(BddFsm fsm, bdd_ptr states);

/* Backward images (for Planning) */
bdd_ptr BddFsm_StrongImageBwd(BddFsm fsm, bdd_ptr states);
bdd_ptr BddFsm_WeakImageBwd(BddFsm fsm, bdd_ptr states);
bdd_ptr BddFsm_MultipleImageBwd(BddFsm fsm, bdd_ptr_set states_set);

/* Belief states images (for Planning) */
bdd_ptr BddFsm_BsImageFwd(BddFsm fsm, bdd_ptr bs);
bdd_ptr BddFsm_BsImageBwd(BddFsm fsm, bdd_ptr bs);

/********************************************************************/

/* BDD cubes */
bdd_ptr BddFsm_GetStateSupport(BddFsm fsm);
bdd_ptr BddFsm_GetNextStateSupport(BddFsm fsm);
bdd_ptr BddFsm_GetInputSupport(BddFsm fsm);

/********************************************************************/

/* Filter functions for applying masks */
bdd_ptr BddFsm_GetMask(BddFsm fsm);
bdd_ptr BddFsm_GetStateMask(BddFsm fsm);
bdd_ptr BddFsm_GetNextStateMask(BddFsm fsm);
bdd_ptr BddFsm_GetInputMask(BddFsm fsm);

bdd_ptr BddFsm_Mask(BddFsm fsm, bdd_ptr bdd);
bdd_ptr BddFsm_MaskStates(BddFsm fsm, bdd_ptr bdd);
bdd_ptr BddFsm_MaskNextStates(BddFsm fsm, bdd_ptr bdd);
bdd_ptr BddFsm_MaskInputs(BddFsm fsm, bdd_ptr bdd);

/********************************************************************/

/* Variables of a model */
/* :TODO: move to an Encoder structure */
node_ptr BddFsm_GetStateVars(BddFsm fsm);
node_ptr BddFsm_GetNextStateVars(BddFsm fsm);
node_ptr BddFsm_GetInputVars(BddFsm fsm);

/* Support, Range, Encoding, and Mask of a variable */
/* :TODO: move to an Encoder structure */
bdd_ptr BddFsm_GetVarSupport(BddFsm fsm, node_ptr var);
node_ptr BddFsm_GetVarRange(BddFsm fsm, node_ptr var);
add_ptr BddFsm_GetVarEncoding(BddFsm fsm, node_ptr var);
bdd_ptr BddFsm_GetVarMask(BddFsm fsm, node_ptr var);

/********************************************************************/

/* Sexpr -> DD */
bdd_ptr BddFsm_EvalCond(BddFsm fsm, node_ptr cond);
add_ptr BddFsm_EvalExpr(BddFsm fsm, node_ptr expr);

/* DD -> Sexpr */
node_ptr BddFsm_WffCond(BddFsm fsm, bdd_ptr bdd);
node_ptr BddFsm_WffExpr(BddFsm fsm, add_ptr add);

/********************************************************************/


int BddFsm_GetNumBddStateVars(BddFsm fsm);
int BddFsm_GetStateLevel(BddFsm fsm);

#endif /* _BDDFSM_H_ */
